/* recVme.c - Record Support Routines for Generic VME records */
/*
 *      Author:         Mark Rivers
 *      Date:           4/9/95
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 */


#include        <vxWorks.h>
#include        <types.h>
#include        <stdioLib.h>
#include        <lstLib.h>
#include        <string.h>

#include        <alarm.h>
#include        <dbDefs.h>
#include        <dbAccess.h>
#include        <dbFldTypes.h>
#include        <devSup.h>
#include        <errMdef.h>
#include        <recSup.h>
#include        <vme.h>
#include        <vmeRecord.h>

/* Create RSET - Record Support Entry Table*/
#define report NULL
#define initialize NULL
static long init_record();
static long process();
#define special NULL
static long get_value();
static long cvt_dbaddr();
static long get_array_info();
static long put_array_info();
#define get_units NULL
#define get_precision NULL
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
#define get_graphic_double NULL
#define get_control_double NULL
#define get_alarm_double NULL

struct rset vmeRSET={
   RSETNUMBER,
   report,
   initialize,
   init_record,
   process,
   special,
   get_value,
   cvt_dbaddr,
   get_array_info,
   put_array_info,
   get_units,
   get_precision,
   get_enum_str,
   get_enum_strs,
   put_enum_str,
   get_graphic_double,
   get_control_double,
   get_alarm_double };

static void monitor();
static long doVmeIo();


static long init_record(pvme, pass)
   struct vmeRecord  *pvme;
   int pass;
{
   if (pass==0) {
      if (pvme->nmax <= 0) pvme->nmax=1;
      pvme->bptr = (char *)calloc(pvme->nmax, sizeof(long));
      pvme->sptr = (char *)calloc(pvme->nmax, sizeof(char));
      return(0);
   }
   return(0);
}

static long process(pvme)
   struct vmeRecord *pvme;
{
   long status;

   status=doVmeIo(pvme); /* Do VME I/O operation */

   recGblGetTimeStamp(pvme);

   /* check event list */
   monitor(pvme);
   /* process the forward scan link record */
   recGblFwdLink(pvme);

   pvme->pact=FALSE;
   return(status);
}

static long get_value(pvme,pvdes)
    struct vmeRecord *pvme;

    struct valueDes *pvdes;
{
   pvdes->no_elements=pvme->nuse;
   pvdes->pvalue = pvme->bptr;
   pvdes->field_type = DBF_LONG;
   return(0);
}

static long cvt_dbaddr(paddr)
struct dbAddr *paddr;
{
   struct vmeRecord *pvme=(struct vmeRecord *)paddr->precord;

   if (paddr->pfield == &(pvme->val)) {
      paddr->pfield = (void *)(pvme->bptr);
      paddr->no_elements = pvme->nmax;
      paddr->field_type = DBF_LONG;
      paddr->field_size = sizeof(long);
      paddr->dbr_field_type = DBF_LONG;
   } else if (paddr->pfield == &(pvme->sarr)) {
      paddr->pfield = (unsigned char *)(pvme->sptr);
      paddr->no_elements = pvme->nmax;
      paddr->field_type = DBF_CHAR;
      paddr->field_size = sizeof(char);
      paddr->dbr_field_type = DBF_CHAR;
   }
   return(0);
}

static long get_array_info(paddr,no_elements,offset)
struct dbAddr *paddr;
long *no_elements;
long *offset;
{
   struct vmeRecord *pvme=(struct vmeRecord *)paddr->precord;

   *no_elements =  pvme->nuse;
   *offset = 0;
   return(0);
}

static long put_array_info(paddr,nNew)
struct dbAddr *paddr;
long nNew;
{
struct vmeRecord *pvme=(struct vmeRecord *)paddr->precord;

   pvme->nuse = nNew;
   if (pvme->nuse > pvme->nmax) pvme->nuse = pvme->nmax;
   return(0);
}


static void monitor(pvme)
    struct vmeRecord   *pvme;
{
   unsigned short monitor_mask;

   /* get previous stat and sevr  and new stat and sevr*/
   monitor_mask = recGblResetAlarms(pvme);
   monitor_mask |= (DBE_LOG|DBE_VALUE);
   if (monitor_mask){
      db_post_events(pvme,pvme->bptr,monitor_mask);
      db_post_events(pvme,pvme->sptr,monitor_mask);
   }
   return;
}

static long doVmeIo(pvme)
   struct vmeRecord   *pvme;
{
   int amod, length, mode, status=ERROR, value, i;
   int *data_array = (int *) pvme->bptr;
   unsigned char *status_array = pvme->sptr;
   char *vme_ptr;
   short word_value;
   char char_value;
   int all_amods[] = {
      VME_AM_SUP_SHORT_IO,
      VME_AM_STD_SUP_DATA,
      VME_AM_EXT_SUP_DATA};
   int all_lengths[] = {1, 2, 4};
   int all_modes[] = {VX_READ, VX_WRITE};

   amod = all_amods[pvme->amod];
   length = all_lengths[pvme->dsiz];
   mode = all_modes[pvme->rdwt];
   sysBusToLocalAdrs(amod, pvme->addr, &vme_ptr);
   if (mode == VX_READ) {
      for (i=0; i<pvme->nuse; i++) {
         switch(length) {
         case 1: status = vxMemProbe( vme_ptr, VX_READ, 1, (char *) &char_value);
                 value = char_value;
                 break;
         case 2: status = vxMemProbe( vme_ptr, VX_READ, 2, (char *) &word_value);
                 value = word_value;
                 break;
         case 4: status = vxMemProbe( vme_ptr, VX_READ, 4, (char *) &value);
                 break;
         }
         data_array[i] = value;
         status_array[i] = status;
         vme_ptr += pvme->ainc;
      }
   } else {               /* VX_WRITE */
      for (i=0; i<pvme->nuse; i++) {
         switch(length) {
         case 1: char_value = data_array[i];
                 status = vxMemProbe( vme_ptr, VX_WRITE, 1, (char *) &char_value);
                 break;
         case 2: word_value = data_array[i];
                 status = vxMemProbe( vme_ptr, VX_WRITE, 2, (char *) &word_value);
                 break;
         case 4: value = data_array[i];
                 status = vxMemProbe( vme_ptr, VX_WRITE, 4, (char *) &value);
                 break;
         }
         vme_ptr += pvme->ainc;
         status_array[i] = status;
      }
   }
   return OK;
}
